#lang scheme

(require "board.ss"
         "test.ss")

(provide find-dests
         find-all-possible)

;; find-roll-dests : dest nat list-of-pos -> hash
;;  Keys of resulting hash are `dest's (and vals are #t)
(define (find-roll-dests init roll occupied)
  
  ;; find-dests : nat hash -> hash
  (define (find-dests roll frontier)
    ;; try : delta delta hash -> hash
    ;;  Accumutate deltas from an old frontier into a new one
    (define (try dr dc so-far)
      (for/fold ([so-far so-far]) ([pos (in-hash-keys frontier)])
        (let ([row (+ (car pos) dr)]
              [col (+ (cadr pos) dc)])
          (let ([c (board-ref row col)])
            (cond
              ;; don't move through a wall:
              [(solid? c) so-far]
              ;; don't move to occupied as last step:
              [(and (= roll 1)
                    (member (list row col) occupied))
               so-far]
              ;; otherwise, ok:
              [else (hash-set so-far (list row col) #t)])))))
    (cond
      [(zero? roll) (add-rooms frontier frontier)]
      [else
       (let* ([new-frontier 
               (try 0 1 (try 1 0 (try -1 0 (try 0 -1 #hash()))))]
              [dests (find-dests (sub1 roll) new-frontier)])
         (add-rooms dests frontier))]))
  
  ;; add-rooms : hash hash -> hash
  ;; Add to `dests' any rooms that we could enter from `frontier':
  (define (add-rooms dests frontier)
    (for/fold ([dests dests]) ([pos (in-hash-keys frontier)])
      (let ([c (board-ref (car pos) (cadr pos))])
        (if (and (door? c) 
                 (not (member pos occupied)) ; check door blocked
                 (not (eq? (car init) (door->room c)))) ; cannot re-enter
            (hash-set dests (list (door->room c) 'V) #t)
            dests))))
  
  ;; remove-all : hash-of-pos list-of-pos -> hash-of-pos
  (define (remove-all frontier occupied)
    (for/fold ([frontier frontier]) ([pos (in-list occupied)])
      (hash-remove frontier pos)))
  
  ;; Start at blank or room:
  (if (number? (car init))
      ;; Initial frontier is the starting position:
      (find-dests roll (hash-set #hash() init #t))
      ;; Initial fronier is all unoccupied doors:
      (find-dests roll (remove-all (find-doors (car init)) occupied))))

(test (find-roll-dests '(7 11) 1 null)
      #hash(((6 11) . #t)
            ((8 11) . #t)
            ((7 10) . #t)
            ((7 12) . #t)))
(test (find-roll-dests '(7 11) 1 '((6 11) (7 10))) ; can land here
      #hash(((8 11) . #t)
            ((7 12) . #t)))
(test (find-roll-dests '(7 11) 2 null)
      #hash(((6 10) . #t)
            ((6 12) . #t)
            ((7 11) . #t)
            ((8 10) . #t)
            ((8 12) . #t)
            ((7 9) . #t)
            ((7 13) . #t)))
(test (find-roll-dests '(7 11) 2 '((6 11) (7 10))) ; can walk here
      #hash(((6 10) . #t)
            ((6 12) . #t)
            ((7 11) . #t)
            ((8 10) . #t)
            ((8 12) . #t)
            ((7 9) . #t)
            ((7 13) . #t)))
(test (find-roll-dests '(7 9) 1 null)
      #hash(((6 9) . #t)
            ((7 8) . #t)
            ((7 10) . #t)
            ((8 9) . #t)
            ((B V) . #t)))
(test (find-roll-dests '(7 9) 2 null)
      #hash(((7 9) . #t)
            ((6 8) . #t)
            ((6 10) . #t)
            ((8 8) . #t)
            ((7 7) . #t)
            ((7 11) . #t)
            ((9 9) . #t)
            ((8 8) . #t)
            ((8 10) . #t)
            ((B V) . #t)))
(test (find-roll-dests '(7 9) 2 '((6 9))) ;; entry blocked
      #hash(((7 9) . #t)
            ((6 8) . #t)
            ((6 10) . #t)
            ((8 8) . #t)
            ((7 7) . #t)
            ((7 11) . #t)
            ((9 9) . #t)
            ((8 8) . #t)
            ((8 10) . #t)))
(test (find-roll-dests '(B V) 1 null)
      #hash(((6 8) . #t)
            ((6 10) . #t)
            ((7 9) . #t)
            ((2 13) . #t)))
(test (find-roll-dests '(B V) 2 null)
      #hash(((6 7) . #t)
            ((6 9) . #t)
            ((7 8) . #t)
            ((6 11) . #t)
            ((7 10) . #t)
            ((8 9) . #t)
            ((1 13) . #t)
            ((3 13) . #t)
            ((L V) . #t)))
(test (find-roll-dests '(B V) 2 '((6 9))) ; lower exit blocked
      #hash(((1 13) . #t)
            ((3 13) . #t)
            ((L V) . #t)))

#;
(time (for* ([row (in-range 25)]
             [col (in-range 25)])
        (find-roll-dests row col 12 null)))

;; find-nonroll-dests : room mode -> hash
;;  Keys of resulting hash are `dests'es (and vals are #t)
(define (find-nonroll-dests room mode)
  (let* ([add (lambda (room dests)
                (hash-set dests (list room 'V) #t))]
         ;; initially empty:
         [dests #hash()]
         ;; staying:
         [dests (if (eq? mode 'S)
                    (add room dests)
                    dests)]
         ;; secret passages:
         [dests (case room
                  [(C) (add 'U dests)]
                  [(U) (add 'C dests)]
                  [(S) (add 'K dests)]
                  [(K) (add 'S dests)]
                  [else dests])])
    dests))

(test (find-nonroll-dests 'B 'V) #hash())
(test (find-nonroll-dests 'B 'S) #hash(((B V) . #t)))
(test (find-nonroll-dests 'C 'V) #hash(((U V) . #t)))
(test (find-nonroll-dests 'C 'S) #hash(((U V) . #t) ((C V) . #t)))

;; find-dests : state nat nat -> hash
;;  Finds all possible desitnations given the player number and 
;;  dice roll (where 0 means not a roll)
(define (find-dests state player roll)
  (let ([init (list-ref state player)])
    (if (zero? roll)
        ;; Didn't roll dice:
        (if (symbol? (car init))
            ;; Try room-to-room
            (find-nonroll-dests (car init) (cadr init))
            ;; Didn't roll when on a blank cell!?
            #hash())
        ;; Rolled the dice:
        (find-roll-dests init roll (filter (lambda (dest) (number? (car dest)))
                                           (remq init state))))))

(test (find-dests '((7 9) (6 9)) 0 2)
      (find-roll-dests '(7 9) 2 '((6 9))))
(test (find-dests '((6 9) (7 9) (U S)) 1 2)
      (find-roll-dests '(7 9) 2 '((6 9))))
(test (find-dests '((6 9) (7 9) (U S)) 1 0)
      #hash())
(test (find-dests '((6 9) (7 9) (U S)) 2 0)
      (find-nonroll-dests 'U 'S))

;; find-all-possible : state nat -> list-of-hash hash
;;  Determines all possibilies for rolling or not rolling
(define (find-all-possible state me)
  (values
   ;; All possible dests for all possible rolls:
   (for/list ([roll (in-range 2 13)])
     (find-dests state me roll))
   ;; All possible dests when not rolling:
   (find-dests state me 0)))
